

# general imports ====
import os, sys
import warnings
import scipy
from scipy.sparse import coo_matrix, issparse
import yaml
import gc
from IPython.utils import io
from pprint import pprint
import time
# import scib
from sklearn.metrics import r2_score
import random
from sklearn import preprocessing
from sklearn.metrics import confusion_matrix
from collections import Counter
import types
import argparse
from argparse import RawTextHelpFormatter
import wandb
from datetime import datetime
import PIL
import PIL.Image
import anndata
import pandas as pd
import scanpy as sc
import gc
# import dask
import squidpy as sq
import numpy as np
import pickle
from scipy import sparse
from scipy.sparse import coo_matrix
import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch_geometric as pyg
from torch_geometric.utils.convert import from_scipy_sparse_matrix
import torchdyn
from torchdyn.core import NeuralODE
import torchcfm
from torchcfm.models import MLP
from torchcfm.utils import plot_trajectories, torch_wrapper
from torch_geometric.loader import NeighborLoader
from tqdm.autonotebook import tqdm
import gdown



# mintflow imports ====


from .. import utils

from .. import generativemodel # exec('import {}.generativemodel'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from .. import modules  # exec('import {}.modules.gnn'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..modules import gnn, neuralODE, mlp, disentonly


from ..modules.impanddisentgl import  MaskLabel #import exec('from {}.modules.impanddisentgl import MaskLabel'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from .. import vardist  # exec('import {}.vardist'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from .. import masking # exec('import {}.masking'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..modules.impanddisentgl import ImputerAndDisentangler # exec('from {}.modules.impanddisentgl import ImputerAndDisentangler'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..modules.disentonly import Disentangler # exec('from {}.modules.disentonly import Disentangler'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..modules.disentonly_twosep import DisentanglerTwoSep # exec('from {}.modules.disentonly_twosep import DisentanglerTwoSep'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..zs_samplers import RandomZSSampler, PerCelltypeZSSampler #exec('from {}.zs_samplers import RandomZSSampler, PerCelltypeZSSampler'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..predadjmat import ListAdjMatPredLoss, AdjMatPredLoss #  exec('from {}.predadjmat import ListAdjMatPredLoss, AdjMatPredLoss'.format(STR_INFLOW_OR_INFLOW_SYNTH))
from ..utils_flowmatching import ModeSampleX0, ModeMinibatchPerm, ModeTimeSched, ModeFMLoss, ConditionalFlowMatcher

from ..modules.cond4flow import Cond4FlowVarphi0

from ..modules.cond4flow_simple3mpls import Cond4FlowVarphi0SimpleMLPs

from ..utils_pyg import PygSTDataGridBatchSampler


from ..evaluation.bioconsv import EvaluatorKmeans, EvaluatorLeiden

from ..evaluation.predxspl import EvalXsplpred, EvalLargeReadoutsXsplpred, EvalOnHVGsXsplpred

from ..modules.gnn_disentangler import GNNDisentangler

from ..kl_annealing import LinearAnnealingSchedule

from ..modules.predictorperCT import PredictorPerCT

from ..utils_multislice import ListSlice, Slice

from ..modules.varphienc4xbar import EncX2Xbar

from ..modules.predictorbatchID import PredictorBatchID


from . import \
    get_defaultconfig_data_train, verify_and_postprocess_config_data_train,\
    get_defaultconfig_data_evaluation, verify_and_postprocess_config_data_evaluation,\
    get_defaultconfig_model, verify_and_postprocess_config_model,\
    get_defaultconfig_training, verify_and_postprocess_config_training,\
    check_listtissue_trtest, base_interface


from .auxiliary_modules import *

#
# from .interface.analresults import disentanglement_jointplot
#
# from .interface.analresults import disentanglement_violinplot

# from . import interface

from ..anneal_decoder_xintxspl import AnnealingDecoderXintXspl


def setup_data(
    dict_all4_configs,
    flag_verbose=True,
    flag_visualise_tissue_sections=True
):
    base_interface.checkif_4configs_are_verified(dict_all4_configs)

    # grab args
    args = types.SimpleNamespace()
    args.flag_verbose = flag_verbose
    args.flag_visualise_tissue_sections = flag_visualise_tissue_sections

    config_data_train, config_data_test, config_model, config_training = \
        dict_all4_configs['config_data_train'], \
        dict_all4_configs['config_data_evaluation'], \
        dict_all4_configs['config_model'], \
        dict_all4_configs['config_training']

    # check if the provided anndata-s share the same gene panel and they all contain count values ===========
    fname_adata0, adata0 = config_data_train[0]['file'], sc.read_h5ad(config_data_train[0]['file'])
    for config_temp in config_data_train + config_data_test:
        if args.flag_verbose:
            print("checking if {} and {} share the same gene panel".format(
                fname_adata0,
                config_temp['file']
            ))

        fname_adata_temp, adata_temp = config_temp['file'], sc.read_h5ad(config_temp['file'])
        if adata_temp.var_names.tolist() != adata0.var_names.tolist():
            raise Exception(
                "Anndata-s {} and {} do not have the same gene panel.".format(
                    fname_adata0,
                    fname_adata_temp
                )
            )

        if not sc._utils.check_nonnegative_integers(adata_temp.X):  # grabbed from https://github.com/scverse/scanpy/blob/0cfd0224f8b0a90317b0f1a61562f62eea2c2927/src/scanpy/preprocessing/_highly_variable_genes.py#L74
            raise Exception(
                "Inflow requires count data, but the anndata in {} seems to have non-count values in adata.X".format(
                    fname_adata_temp
                )
            )
        else:
            if args.flag_verbose:
                print("    >>> also checked that anndata.X contains count data.")

        del fname_adata_temp, adata_temp
        gc.collect()

    del fname_adata0, adata0, config_temp
    gc.collect()

    # set device ===
    if config_training['flag_use_GPU']:
        if torch.cuda.is_available():
            device = torch.device("cuda:0")
        else:
            print("Although flag_use_GPU is set True in {}, but cuda is not available --> falling back to CPU.".format(
                args.file_config_training
            ))
            device = torch.device("cpu")
    else:
        device = torch.device("cpu")

    if args.flag_verbose:
        print("\n\nDevice is set to {}.\n\n".format(device))

    # Create list tissue training =========

    def _convert_TrueFalse_to_bool(dict_input):
        '''
        Given an input dictionary, if any element is "True" or "Fasle" it converts them to booleans.
        :param dict_input:
        :return: the modified dictionary.
        '''
        for k in dict_input:
            if isinstance(dict_input[k], bool):
                raise Exception(
                    'Found {} of type bool. Did you write "true" or "flase" (or True or False without double-quotation-s) instead of "True" or "False" in any of the config files? If so, please modify to latter.'.format(
                        k
                    )
                )

            if dict_input[k] in ['True', 'False']:
                dict_input[k] = (dict_input[k] == 'True')

        return dict_input


    list_slice = []
    for dict_current_anndata in config_data_train:
        current_anndata = sc.read_h5ad(dict_current_anndata['file'])
        unnorm_anndata = current_anndata.copy()
        sc.pp.normalize_total(
            current_anndata,
            target_sum=config_training['val_scppnorm_total'],
            inplace=True
        )

        list_slice.append(
            Slice(
                adata=current_anndata,
                adata_before_scppnormalize_total=unnorm_anndata,
                dict_obskey={
                    'cell_type':dict_current_anndata['obskey_cell_type'],
                    'sliceid_to_checkUnique':dict_current_anndata['obskey_sliceid_to_checkUnique'],
                    'x':dict_current_anndata['obskey_x'],
                    'y':dict_current_anndata['obskey_y'],
                    'biological_batch_key':dict_current_anndata['obskey_biological_batch_key']
                },
                kwargs_compute_graph={
                    'spatial_key': 'spatial',
                    'library_key': None,
                     **_convert_TrueFalse_to_bool(dict_current_anndata['config_neighbourhood_graph'])
                },
                flag_use_custompygsampler=True,
                kwargs_pygdl_train={
                    'num_neighbors': [dict_current_anndata['config_dataloader_train']['num_neighbors']] * config_model['num_graph_hops'],
                    'width_window': dict_current_anndata['config_dataloader_train']['width_window'],
                    'min_numpoints_ingrid': dict_current_anndata['config_dataloader_train']['min_numpoints_ingrid'],
                    'flag_disable_randoffset': False
                },
                kwargs_pygdl_test={
                    'num_neighbors': [-1] * config_model['num_graph_hops'],
                    'width_window': dict_current_anndata['config_dataloader_train']['width_window'],
                    'min_numpoints_ingrid': 1,
                    'flag_disable_randoffset': True
                },
                neighgraph_num_hops_computeNCC=1,
                kwargs_sq_pl_spatial_scatter=_convert_TrueFalse_to_bool(dict_current_anndata['config_sq_pl_spatial_scatter']),
                device=device,
                batchsize_compute_NCC=dict_current_anndata['batchsize_compute_NCC']
            )
        )


    list_slice = ListSlice(
        list_slice=list_slice
    )

    if args.flag_verbose:
        print("\n\ncreated list_slice for training.")
        for sl in list_slice.list_slice:
            print("Tissue {} --> {} cells".format(
                set(sl.adata.obs[sl.dict_obskey['sliceid_to_checkUnique']]),
                sl.adata.shape[0]
            ))
        print("\n\n")


    # create test_list_slice for evaluation ===
    test_list_slice = []
    for dict_current_anndata in config_data_test:
        current_anndata = sc.read_h5ad(dict_current_anndata['file'])
        unnorm_anndata = current_anndata.copy()
        sc.pp.normalize_total(
            current_anndata,
            target_sum=config_training['val_scppnorm_total'],
            inplace=True
        )

        test_list_slice.append(
            Slice(
                adata=current_anndata,
                adata_before_scppnormalize_total=unnorm_anndata,
                dict_obskey={
                    'cell_type':dict_current_anndata['obskey_cell_type'],
                    'sliceid_to_checkUnique':dict_current_anndata['obskey_sliceid_to_checkUnique'],
                    'x':dict_current_anndata['obskey_x'],
                    'y':dict_current_anndata['obskey_y'],
                    'biological_batch_key':dict_current_anndata['obskey_biological_batch_key']
                },
                kwargs_compute_graph={
                    'spatial_key': 'spatial',
                    'library_key': None,
                     **_convert_TrueFalse_to_bool(dict_current_anndata['config_neighbourhood_graph'])
                },
                flag_use_custompygsampler=True,
                kwargs_pygdl_train={
                    'num_neighbors': [dict_current_anndata['config_dataloader_test']['num_neighbors']] * config_model['num_graph_hops'],
                    'width_window': dict_current_anndata['config_dataloader_test']['width_window'],
                    'min_numpoints_ingrid': dict_current_anndata['config_dataloader_test']['min_numpoints_ingrid'],
                    'flag_disable_randoffset': False
                },  # dummy training dl's which are never used.
                kwargs_pygdl_test={
                    'num_neighbors': [-1] * config_model['num_graph_hops'],
                    'width_window': dict_current_anndata['config_dataloader_test']['width_window'],
                    'min_numpoints_ingrid': 1,
                    'flag_disable_randoffset': True
                },
                neighgraph_num_hops_computeNCC=1,
                kwargs_sq_pl_spatial_scatter=_convert_TrueFalse_to_bool(dict_current_anndata['config_sq_pl_spatial_scatter']),
                device=device,
                batchsize_compute_NCC=dict_current_anndata['batchsize_compute_NCC']
            )
        )

    test_list_slice = ListSlice(
        list_slice=test_list_slice,
        prev_list_slice_to_imitate=list_slice
    )

    check_listtissue_trtest.check(
        train_list_slice=list_slice,
        test_list_slice=test_list_slice
    )

    if args.flag_verbose:
        print("\n\ncreated list_slice for evaluation.")
        for sl in test_list_slice.list_slice:
            print("Tissue {} --> {} cells".format(
                set(sl.adata.obs[sl.dict_obskey['sliceid_to_checkUnique']]),
                sl.adata.shape[0]
            ))
        print("\n\n")



    # show the neighbourhood graphs ======
    if args.flag_visualise_tissue_sections:
        # Train
        list_slice.show_scatters_4cli(path_output=None, flag_dump_output=False)

        # Test
        test_list_slice.show_scatters_4cli(path_output=None, flag_dump_output=False)


    if args.flag_verbose:
        print("\n\n\n")
        print("The provided cell types are aggregated/mapped to mintflow cell types as follow:")
        pprint(list_slice.map_CT_to_inflowCT)
        print("\n\n")

    if args.flag_verbose:
        print("\n\n\n")
        print("The provided biological batch IDs are aggregated/mapped to mintflow batch IDs as follows")
        pprint(list_slice.map_Batchname_to_inflowBatchID)
        print("\n\n")


    if args.flag_verbose:
        with torch.no_grad():
            print("One-hot encoded batch ID for each sample (tissue):")
            for sl in list_slice.list_slice:
                print(
                    "     sample {} --> batch ID {}".format(
                        list(sl.adata.obs[sl.dict_obskey['sliceid_to_checkUnique']])[0],
                        set(sl.ten_BatchEmb.argmax(1).tolist())
                    )
                )



    # compute/report the maximum number of subgraphs
    print("Customised neighbourloader sampler: computing some initial stats (max number of central nodes, etc) for each tissue.")
    list_maxsize_subgraph = []
    dict_slideID_to_maxnumcentralnodes, dict_slideID_to_worsecasebatchsize = {}, {}
    for idx_sl, sl in enumerate(list_slice.list_slice):
        print("Tissue # {}".format(idx_sl+1))
        sl: Slice
        pyg_neighloader_train, pyg_neighloader_test = sl.pyg_dl_train, sl.pyg_dl_test

        if pyg_neighloader_train.batch_size is None:
            batch_size = pyg_neighloader_train.batch_sampler.get_maxnumpoints_insquare()
            maxsize_subgraph = 10 + np.sum(
                np.cumprod([batch_size] + pyg_neighloader_train.node_sampler.num_neighbors.values)
            )
            maxsize_subgraph = int(maxsize_subgraph)

        else:
            maxsize_subgraph = 10 + np.sum(
                np.cumprod([pyg_neighloader_train.batch_size] + pyg_neighloader_train.node_sampler.num_neighbors.values))
            maxsize_subgraph = int(maxsize_subgraph)

            print("maxsize_subgraph is equal to {}. If it's too big, try reducing width_window.".format(
                maxsize_subgraph
            ))

        list_maxsize_subgraph.append(maxsize_subgraph)

        if pyg_neighloader_train.batch_size is None:
            print("    width_window={} --> [maxnum_centralnodes:{},    worse-case batchsize:{}]".format(
                sl.kwargs_pygdl_train['width_window'],
                pyg_neighloader_train.batch_sampler.get_maxnumpoints_insquare(),
                maxsize_subgraph
            ))

            dict_slideID_to_maxnumcentralnodes[list(sl.adata.obs[sl.dict_obskey['sliceid_to_checkUnique']])[0]] = pyg_neighloader_train.batch_sampler.get_maxnumpoints_insquare()
            dict_slideID_to_worsecasebatchsize[list(sl.adata.obs[sl.dict_obskey['sliceid_to_checkUnique']])[0]] = maxsize_subgraph


    maxsize_subgraph = max(list_maxsize_subgraph)



    # plot the crops that show window sizes =========
    if args.flag_visualise_tissue_sections:
        # Train
        list_slice.show_pygbatch_windows_4cli(
            path_output=None,
            str_train_or_test='train',
            dict_slideID_to_maxnumcentralnodes=dict_slideID_to_maxnumcentralnodes,
            dict_slideID_to_worsecasebatchsize=dict_slideID_to_worsecasebatchsize,
            flag_dump_output=False
        )

        # Test
        test_list_slice.show_pygbatch_windows_4cli(
            path_output=None,
            str_train_or_test='test',
            dict_slideID_to_maxnumcentralnodes=None,
            dict_slideID_to_worsecasebatchsize=None,
            flag_dump_output=False
        )


    return dict(
        train_list_tissue_section=list_slice,
        evaluation_list_tissue_section=test_list_slice,
        maxsize_subgraph=maxsize_subgraph
    )




